% Eickmeier/Lemke/Marcellino (2013)
% Code for TV-FAVAR analysis
% Last modification: 28 May 2014

% LAst change (28 May 2014) also explicitly imposes stability condition on
% the smoothing algorithm, i.e. a smoothing step (recall: going backward
% through the sample using filtered and predicted states as input) is also
% applied. In the previous version, the results for the 'RESTRICTED'smoother were
% obtained by running the normal smoother taken the results on the
% restricted filter and predictor as given. However, it turned out that this
% did not suffice to have stability (or more generally: satisfy the
% restictions) for the smoothed estimates as well at all points in time.
% Therefore, now in the new version, the restriction is applied also in the smoothing step!

% joint Kalman filter for GROUP of S time-const or timevar. SSMs
% The KF is run of each of the S  SSMs, but at each t, a joint restriction
% is checked via the binary-valued  function RestrFuncStr (a string, i.e. the
% function name). If restriction is jointly satisfied by all the
% filtered states at t, go on with usual KF, otherwise no  updating is done,
% but the previous (t-1) filtered state is taken.
function [afis,Pfis,yprs,Fprs,aprs,Pprs,llh,RestrSatTS,asmos,Psmos,epsSmo,etaSmo,VarepsSmo,VaretaSmo,rSmo,NSmo,uSmo,DSmo]=WL_KalmanGroupRestr(yt,dt,Zt,Ht,ct,Kapt,Rt,Qt,a0,P0,inista,istv,RestrFuncStr);


% The input and output is comparable to  WL_Kalman except
% that all KF inputs and outputs are cells of dimension Sx1, where each cell element contains  the matrix of the respective SSM
% (have to use cell and not array since dimension can always change)
% exception (this version):
% inista,istv are scalars, where it is assumed that inista (indicating what type of  state initialization) and istv (tv or not SSM) are the same for all S SSMs

% Additional (wrt WL_Kalman) Input:
% - RestrFuncStr: String, the name of the function that takes as input the S different state estimates at time t and checks for the joint restriction
% - ...

% Additional Output:
% - RestrSatTS: Tx1 vector indicating (1) when the restriction was satisfied (and hence usual KF updating was done) and when not (0)
% - ...


%*******************
%   Know what the restiction-check function is
%*******************
RestrFunc=str2func(RestrFuncStr);

%*******************
%   Dimensions
%*******************
S=size(yt,1);
Tbig=size(yt{1},1); % assumed to be the same for all S SSMs
nmeas=ones(S,1);
nsta=ones(S,1);
for ss=1:S;
    nmeas(ss)=size(yt{ss},2);
    nsta(ss)=size(Kapt{ss},1);
end;

RestrSatTS=zeros(Tbig,1);
RestrSatTS_smo=zeros(Tbig,1);

%*******************
%   Provide S-dimensional cells for empty result matrices and multidim
%   arrays
%*******************
afis=cell(S,1);
Pfis=cell(S,1);
aprs=cell(S,1);
Pprs=cell(S,1);
yprs=cell(S,1);
vprs=cell(S,1);
Fprs=cell(S,1);
llh=cell(S,1);

epsSmo=cell(S,1);
etaSmo=cell(S,1);
VarepsSmo=cell(S,1);
VaretaSmo=cell(S,1);
rSmo=cell(S,1);
NSmo=cell(S,1);
uSmo=cell(S,1);
DSmo=cell(S,1);

asmos=afis;
Psmos=Pfis;

%Make empty containers for capturing the filtered and smoothed estimates
% these are cells capturing respective results for each of the S state space models
for ss=1:S;
    afis{ss}=NaN*ones(Tbig,nsta(ss));
    Pfis{ss}=NaN*ones(nsta(ss),nsta(ss),Tbig);
    aprs{ss}=NaN*ones(Tbig,nsta(ss));
    Pprs{ss}=NaN*ones(nsta(ss),nsta(ss),Tbig);
    yprs{ss}=NaN*ones(Tbig,nmeas(ss));
    vprs{ss}=yprs{ss};
    Fprs{ss}=NaN*ones(nmeas(ss),nmeas(ss),Tbig);
    llh{ss}=-nmeas(ss)*Tbig/2*log(2*pi);
    
    asmos{ss}=NaN*ones(Tbig,nsta(ss));
    Psmos{ss}=NaN*ones(nsta(ss),nsta(ss),Tbig);
end;

%*******************
%   Initialization
%*******************
afiLoc=cell(S,1);
PfiLoc=cell(S,1);
if inista==1;
    for  ss=1:S;
        afiLoc{ss}=a0{ss};
        PfiLoc{ss}=P0{ss};
    end;
elseif inista==0;
    a0=cell(S,1);
    P0=cell(S,1);
    for ss=1:S;
        % only works  for stationary non-tv transition eq
        a0{ss}=inv(eye(nsta(ss))-Kapt{ss})*ct{ss};
        vecP0Loc=inv(eye(nsta(ss)^2)-kron(Kapt{ss},Kapt{ss}))*vec(Rt{ss}*Qt{ss}*Rt{ss}');
        P0{ss}=reshape(vecP0Loc,nsta(ss),nsta(ss));
        afiLoc{ss}=a0;
        PfiLoc{ss}=P0;
    end;
end;
checkRestrInd=RestrFunc(afiLoc);
if checkRestrInd==0;
    error('WL: The set of initial states does not satisfy the joint restriction!');
else; % NOTE: this is a very big else!
    
%*******************
% Group of Kalman filters, time-indep SSM
%*******************
if istv==0;
    % if time-indep SSM, can set syst matrices once and for all
    Kap=cell(S,1);
    R=cell(S,1);
    c=cell(S,1);
    Q=cell(S,1);
    Z=cell(S,1);
    d=cell(S,1);
    H=cell(S,1);
    for ss=1:S;
        Kap{ss,1}=Kapt{ss};
        R{ss,1}=Rt{ss};
        c{ss,1}=ct{ss};
        Q{ss,1}=Qt{ss};
        Z{ss,1}=Zt{ss};
        d{ss,1}=dt{ss};
        H{ss,1}=Ht{ss};
    end;
    %----------------------------------
    % Run the restricted filter
    %----------------------------------
    for tt=1:Tbig;
        % NOTE: afiLoc and PfiLoc are cells, whereas aprLoc and other ...prLoc
        % are arrays or matrices.

        afiPropAll=cell(S,1); %'Prop' stands for Proposal
        PfiPropAll=cell(S,1);

        for ss=1:S;
            % Prediction step
            aprLoc=Kap{ss}*afiLoc{ss}+c{ss};
            PprLoc=Kap{ss}*PfiLoc{ss}*Kap{ss}'+R{ss}*Q{ss}*R{ss}';
            yprLoc=Z{ss}*aprLoc+d{ss};
            FprLoc=Z{ss}*PprLoc*Z{ss}'+H{ss};
            if rcond(FprLoc) > 10^-12
                invFprLoc=inv(FprLoc);
            else
                invFprLoc=pinv(FprLoc);
                %dummy=1;
            end
            vprLoc=yt{ss}(tt,:)'-yprLoc;
            % Note: all these pr variables can be stored anyway (we do not check
            % their compliance with the restrictions).
            % 28/05/2014: note also that in the RW case (Kap=I), the prediction
            % satisfies the restriction iff the filtered one satisfies it!

            aprs{ss}(tt,:)=aprLoc';
            Pprs{ss}(:,:,tt)=PprLoc;
            yprs{ss}(tt,:)=yprLoc';
            vprs{ss}(tt,:)=vprLoc';
            Fprs{ss}(:,:,tt)=FprLoc;

            % Updating step ("Propose" that to the stationarity check)
            afiProp=aprLoc+PprLoc*Z{ss}'*invFprLoc*(yt{ss}(tt,:)'-yprLoc);
            PfiProp=PprLoc-PprLoc*Z{ss}'*invFprLoc*Z{ss}*PprLoc;
            % log likelihood
            llh{ss}=llh{ss}-0.5*log(det(FprLoc))-0.5*vprLoc'*invFprLoc*vprLoc;

            afiPropAll{ss}=afiProp;
            PfiPropAll{ss}=PfiProp;
        end; % ss=1:S;
        checkRestrInd=RestrFunc(afiPropaAll);
        if checkRestrInd==1;
            RestrSatTS(tt)=1;
            afiLoc=afiPropAll;
            PfiLoc=PfiPropAll;
            %else: afiLoc and PfiLoc remain on previous period's value
        end;
        for ss=1:S;
            afis{ss}(tt,:)=afiLoc{ss}';
            Pfis{ss}(:,:,tt)=PfiLoc{ss};
        end;
    end;
    
    %----------------------------------
    % -- OUTDATED, see smoother below --- take the restricted filtered path and apply smoothing
    %----------------------------------
    % for ss=1:S;
        % 28/05/2014. THE FOLLOWING three lines are replaced by running
        % the restricted smoother, see loop below,(like the filter above)
        % MOREOVER, we use a nested version for the smoother that applies
        % to both the istv==0 and istv==1 case
        
        %[asms,Psms]=WL_Kalmansmo(afis{ss},Pfis{ss},aprs{ss},Pprs{ss},Kapt{ss},istv);
        %asmos{ss}=asms;
        %Psmos{ss}=Psms;

        % We completely skip the computation of smoothed residuals, (i.e.
        % programme should return NA) urrently they apparently are not used by other parts
        % of the code suite?!

        %         [epsSmo_ss,etaSmo_ss,VarepsSmo_ss,VaretaSmo_ss,rSmo_ss,NSmo_ss,uSmo_ss,DSmo_ss]=WL_Kalmandistsmo(Pprs{ss},yprs{ss},Fprs{ss},Kapt{ss},Zt{ss},Ht{ss},Rt{ss},Qt{ss},yt{ss},istv);
        %         epsSmo{ss}=epsSmo_ss;
        %         etaSmo{ss}=etaSmo_ss;
        %         VarepsSmo{ss}=VarepsSmo_ss;
        %         VaretaSmo{ss}=VaretaSmo_ss;
        %         rSmo{ss}=rSmo_ss;
        %         NSmo{ss}=NSmo_ss;
        %         uSmo{ss}=uSmo_ss;
        %         DSmo{ss}=DSmo_ss;
    % end;
end; %if istv==0;

%*******************
% Group of Kalman filters, time-dep SSM
%*******************
if istv==1;
    for tt=1:Tbig;
        % Set tt-current system matrices
        Kap=cell(S,1);
        R=cell(S,1);
        c=cell(S,1);
        Q=cell(S,1);
        Z=cell(S,1);
        d=cell(S,1);
        H=cell(S,1);
        for ss=1:S;
            Kap{ss,1}=Kapt{ss}(:,:,tt);
            R{ss,1}=Rt{ss}(:,:,tt);
            c{ss,1}=ct{ss}(tt,:)';
            Q{ss,1}=Qt{ss}(:,:,tt);
            Z{ss,1}=Zt{ss}(:,:,tt);
            d{ss,1}=dt{ss}(tt,:)';
            H{ss,1}=Ht{ss}(:,:,tt);
        end;

        % NOTE: afiLoc and PfiLoc are cells, whereas aprLoc and other ...prLoc
        % are arrays or matrices.

        afiPropAll=cell(S,1);
        PfiPropAll=cell(S,1);

        for ss=1:S;
            % Prediction step
            aprLoc=Kap{ss}*afiLoc{ss}+c{ss};
            PprLoc=Kap{ss}*PfiLoc{ss}*Kap{ss}'+R{ss}*Q{ss}*R{ss}';
            yprLoc=Z{ss}*aprLoc+d{ss};
            FprLoc=Z{ss}*PprLoc*Z{ss}'+H{ss};
            if rcond(FprLoc) > 10^-12
                invFprLoc=inv(FprLoc);
            else
                invFprLoc=pinv(FprLoc);
                %dummy=1;
            end
            vprLoc=yt{ss}(tt,:)'-yprLoc;
            % Note: all these pr variables can be stored anyway (we do not check
            % their compliance with the restrictions.
            aprs{ss}(tt,:)=aprLoc';
            Pprs{ss}(:,:,tt)=PprLoc;
            yprs{ss}(tt,:)=yprLoc';
            vprs{ss}(tt,:)=vprLoc';
            Fprs{ss}(:,:,tt)=FprLoc;

            % Updating step
            afiProb=aprLoc+PprLoc*Z{ss}'*invFprLoc*(yt{ss}(tt,:)'-yprLoc);
            PfiProb=PprLoc-PprLoc*Z{ss}'*invFprLoc*Z{ss}*PprLoc;
            % log likelihood
            llh{ss}=llh{ss}-0.5*log(det(FprLoc))-0.5*vprLoc'*invFprLoc*vprLoc;

            afiPropAll{ss}=afiProb;
            PfiPropAll{ss}=PfiProb;
        end; % ss=1:S;
        checkRestrInd=RestrFunc(afiPropAll);
        if checkRestrInd==1;
            RestrSatTS(tt)=1;
            afiLoc=afiPropAll;
            PfiLoc=PfiPropAll;
            %else: afiLoc and PfiLoc remain on previous period's value
        end;
        for ss=1:S;
            afis{ss}(tt,:)=afiLoc{ss}';
            Pfis{ss}(:,:,tt)=PfiLoc{ss};
        end;
    end;
    %     % take the restricted filtered path and apply smoothing
    % SEE COMMENT ABOVE IN THE istv==0 case: SMOOTHING IS DONE BELOW NOW
    %     for ss=1:S;
    %         [asms,Psms]=WL_Kalmansmo(afis{ss},Pfis{ss},aprs{ss},Pprs{ss},Kapt{ss},istv);
    %         [epsSmo_ss,etaSmo_ss,VarepsSmo_ss,VaretaSmo_ss,rSmo_ss,NSmo_ss,uSmo_ss,DSmo_ss]=WL_Kalmandistsmo(Pprs{ss},yprs{ss},Fprs{ss},Kapt{ss},Zt{ss},Ht{ss},Rt{ss},Qt{ss},yt{ss},istv);
    % 
    %         asmos{ss}=asms;
    %         Psmos{ss}=Psms;
    % 
    %         epsSmo{ss}=epsSmo_ss;
    %         etaSmo{ss}=etaSmo_ss;
    %         VarepsSmo{ss}=VarepsSmo_ss;
    %         VaretaSmo{ss}=VaretaSmo_ss;
    %         rSmo{ss}=rSmo_ss;
    %         NSmo{ss}=NSmo_ss;
    %         uSmo{ss}=uSmo_ss;
    %         DSmo{ss}=DSmo_ss;
    %     end;
end;
%*******************
% Take the restricted filtered path and apply restricted smoothing.
% This version works for istv==0 and istv==1 (need to only take care of
% Kappa matrix)
%*******************
%if non-tv version applies, make a sequence of Kappas
if istv==0;
    Kapt_input=Kapt; % this is how it was given to the function as cell with S fixed entries
    % now turn it to a cell with S arrays
    Kapt=cell(S,1);
    for ss=1:S;
        KapArray=NaN(nsta(ss),nsta(ss),Tbig);
        for tt=1:Tbig;
            KapArray(:,:,tt)=Kapt_input{ss};
        end;
        Kapt{ss}=KapArray;
    end;
end;

asmLoc=cell(S,1);
PsmLoc=cell(S,1);
% Now work backwards
% for tt=Tbig:
for ss=1:S
    asmos{ss}(Tbig,:)=afis{ss}(Tbig,:); % note this satisfies the resriction as it comes from the filter that ran under restriction
    Psmos{ss}(:,:,Tbig)=Pfis{ss}(:,:,Tbig);
    asmLoc{ss}=afis{ss}(Tbig,:)';
    PsmLoc{ss}=Pfis{ss}(:,:,Tbig);
end;
for tt=Tbig-1:(-1):1; %run backwards
asmPropAll=cell(S,1); %'Prop' stands for Proposal (to be checked against the restrictions to be satisfied)
PsmPropAll=cell(S,1);
for ss=1:S;
    %retrieve the results for ss and tt for filter and predictor, which are
    %required input to the smoother
    afiLoc=afis{ss}(tt,:)';
    PfiLoc=Pfis{ss}(:,:,tt);
    aprLoc=aprs{ss}(tt+1,:)'; % NOTE THE TIMING! (the prs are stored as predictions FOR t, here the prediction AT t is needed)
    PprLoc=Pprs{ss}(:,:,tt+1)'; % NOTE THE TIMING!
    % compute the smoother (proposal)
    if rcond(PprLoc) > 10^-12
        invPprLoc=inv(PprLoc);
    else
        invPprLoc=pinv(PprLoc);
        %dummy=1;
    end
    KapLoc=Kapt{ss}(:,:,tt);
    PstarLoc=PfiLoc*KapLoc'*invPprLoc;
    asmProp=afiLoc+PstarLoc*(asmLoc{ss}-aprLoc);
    PsmProp=PfiLoc+PstarLoc*(PsmLoc{ss}-PprLoc)*PstarLoc';

    asmPropAll{ss}=asmProp;
    PsmPropAll{ss}=PsmProp;
end; % ss=1:S;
checkRestrInd=RestrFunc(asmPropAll);
if checkRestrInd==1;
    RestrSatTS_smo(tt)=1;
    asmLoc=asmPropAll; % this cell going then as input to the next period
    PsmLoc=PsmPropAll;
    %else: afiLoc and PfiLoc remain on previous period's value
end;
    for ss=1:S;
        asmos{ss}(tt,:)=asmLoc{ss}'; %so if asmLoc was assigned the updated figure, this is saved for period tt, o/w it is the old one that is taken over 
        Psmos{ss}(:,:,tt)=PsmLoc{ss};
    end;
end;

    
end; % initial  checkRestrInd is not 0;